@@ -6,7 +6,7 @@
 #include "../elfchk/hdr.h"
 #include "../elfchk/proghdr.h"
 
-ElfFile *check_file(char *filename, struct tetj_handle *journal, int isProgram);
-void check_lib(char *filename, struct tetj_handle *journal, int isProgram);
+int check_file(ElfFile *elffile, struct tetj_handle *journal, Elf_type isLib);
+void check_lib(ElfFile *elffile, struct tetj_handle *journal, Elf_type isLib, int modules);
 
 #endif /* __CHECK_FILE_H */
