@@ -9,6 +9,7 @@
 #include <stdio.h>
 #include <unistd.h>
 #include <string.h>
+#include <malloc.h>
 #include "elfchk.h"
 #include "dwarf.h"
 
@@ -74,12 +75,33 @@ decode_sleb128 (data, length_return)
   return result;
 }
 
+int encoded_ptr_size (int encoding) {
+   int size = 0;
+   switch (encoding & 0x7) {
+      case 0: 
+	       size = sizeof(void*);
+		   break;
+      case DW_EH_PE_udata2: 
+	       size = 2; 
+		   break;
+      case DW_EH_PE_udata4: 
+	       size = 4; 
+		   break;
+      case DW_EH_PE_udata8:
+	       size = 8; 
+		   break;
+      default:	
+	       break;
+   }
+   
+   return size;
+}
 
-
-int check_CFI(unsigned char *ptr, int *error)
+int check_CFI(unsigned char *ptr, int *error, int ptr_encoding)
 {
         int	numused,used=0;
         int	tmp;
+		long tmpaddr;
         int     opcode;
         unsigned char     rawop;
 /* XXXSTU - Need to check the values read in each record */
@@ -111,18 +133,21 @@ dumpbytes(ptr,8);
 	 */
 	case DW_CFA_advance_loc:
 		if (elfchk_debug & DEBUG_DWARF_CONTENTS) {
-			fprintf(stderr,"DW_CFA_advance_loc\n" );
+			fprintf(stderr,"DW_CFA_advance_loc %x\n",rawop & 0x3F );
 		}
 		break;
 
 	case DW_CFA_offset:
 		if (elfchk_debug & DEBUG_DWARF_CONTENTS) {
-			fprintf(stderr,"DW_CFA_offset\n" );
+			fprintf(stderr,"DW_CFA_offset" );
 		}
 		/* Operand 1 - ULEB128 register */
 		tmp = decode_uleb128(ptr,&numused);
                 ptr+=numused;
                 used+=numused;
+		if (elfchk_debug&DEBUG_DWARF_CONTENTS) {
+			fprintf(stderr," r%x\n", tmp );
+                }
 		break;
 
 	case DW_CFA_restore:
@@ -143,9 +168,15 @@ dumpbytes(ptr,8);
                 }
 
 		/* Operand 1 - address */
-		tmp = *(int *)(ptr);
-                ptr += sizeof(int);
-                used += sizeof(int);
+		if (ptr_encoding) {
+		 tmpaddr = (long)read_FDE_encoded(ptr, ptr_encoding, &numused);
+		 ptr += numused;
+		 used += numused;
+		} else {
+            tmp = *(long *)(ptr);
+            ptr += sizeof(long);
+            used += sizeof(long);
+		}
 		break;
 
 	case DW_CFA_advance_loc1:
@@ -190,11 +221,17 @@ dumpbytes(ptr,8);
 		tmp = decode_uleb128(ptr,&numused);
                 ptr+=numused;
                 used+=numused;
+		if (elfchk_debug&DEBUG_DWARF_CONTENTS) {
+			fprintf(stderr," r%d", tmp );
+                }
 
 		/* Operand 2 - ULEB128 offset */
 		tmp = decode_uleb128(ptr,&numused);
                 ptr += numused;
                 used += numused;
+		if (elfchk_debug&DEBUG_DWARF_CONTENTS) {
+			fprintf(stderr," offset %d\n", tmp );
+                }
 		break;
 
 	case DW_CFA_restore_extended:
@@ -217,6 +254,9 @@ dumpbytes(ptr,8);
 		tmp = decode_uleb128(ptr,&numused);
                 ptr+=numused;
                 used+=numused;
+		if (elfchk_debug&DEBUG_DWARF_CONTENTS) {
+			fprintf(stderr," r%d\n", tmp );
+                }
 		break;
 
 	case DW_CFA_same_value:
@@ -228,6 +268,9 @@ dumpbytes(ptr,8);
 		tmp = decode_uleb128(ptr,&numused);
                 ptr+=numused;
                 used+=numused;
+		if (elfchk_debug&DEBUG_DWARF_CONTENTS) {
+			fprintf(stderr," r%d\n", tmp );
+                }
 		break;
 
 	case DW_CFA_register:
@@ -239,11 +282,17 @@ dumpbytes(ptr,8);
 		tmp = decode_uleb128(ptr,&numused);
                 ptr+=numused;
                 used+=numused;
+		if (elfchk_debug&DEBUG_DWARF_CONTENTS) {
+			fprintf(stderr," r%d", tmp );
+                }
 
 		/* Operand 2 - ULEB128 register */
 		tmp = decode_uleb128(ptr,&numused);
                 ptr += numused;
                 used += numused;
+		if (elfchk_debug&DEBUG_DWARF_CONTENTS) {
+			fprintf(stderr," r%d\n", tmp );
+                }
 		break;
 
 	case DW_CFA_remember_state:
@@ -260,18 +309,28 @@ dumpbytes(ptr,8);
 
 	case DW_CFA_def_cfa:
 		if (elfchk_debug&DEBUG_DWARF_CONTENTS) {
-			fprintf(stderr,"DW_CFA_def_cfa\n" );
+			fprintf(stderr,"DW_CFA_def_cfa" );
                 }
 
 		/* Operand 1 - ULEB128 register */
 		tmp = decode_uleb128(ptr,&numused);
                 ptr += numused;
                 used += numused;
+		if (elfchk_debug&DEBUG_DWARF_CONTENTS) {
+			fprintf(stderr," r%x", tmp );
+                }
 
 		/* Operand 2 - ULEB128 offset */
 		tmp = decode_uleb128(ptr,&numused);
                 ptr += numused;
                 used += numused;
+		if (elfchk_debug&DEBUG_DWARF_CONTENTS) {
+			fprintf(stderr," offset %x", tmp );
+                }
+
+		if (elfchk_debug&DEBUG_DWARF_CONTENTS) {
+			fprintf(stderr,"\n" );
+                }
 		break;
 
 	case DW_CFA_def_cfa_register:
@@ -283,17 +342,23 @@ dumpbytes(ptr,8);
 		tmp = decode_uleb128(ptr,&numused);
                 ptr += numused;
                 used += numused;
+		if (elfchk_debug&DEBUG_DWARF_CONTENTS) {
+			fprintf(stderr," r%d\n", tmp );
+                }
 		break;
 
 	case DW_CFA_def_cfa_offset:
 		if (elfchk_debug&DEBUG_DWARF_CONTENTS) {
-			fprintf(stderr,"DW_CFA_def_cfa_offset\n" );
+			fprintf(stderr,"DW_CFA_def_cfa_offset" );
                 }
 
 		/* Operand 1 - ULEB128 register */
 		tmp = decode_uleb128(ptr,&numused);
                 ptr += numused;
                 used += numused;
+		if (elfchk_debug&DEBUG_DWARF_CONTENTS) {
+			fprintf(stderr," %d\n", tmp );
+                }
 		break;
 
 	case DW_CFA_GNU_args_size:
@@ -305,6 +370,9 @@ dumpbytes(ptr,8);
 		tmp = decode_uleb128(ptr,&numused);
                 ptr += numused;
                 used += numused;
+		if (elfchk_debug&DEBUG_DWARF_CONTENTS) {
+			fprintf(stderr," r%d\n", tmp );
+                }
 		break;
 
 	case DW_CFA_GNU_negative_offset_extended:
@@ -316,11 +384,17 @@ dumpbytes(ptr,8);
 		tmp = decode_uleb128(ptr,&numused);
                 ptr += numused;
                 used += numused;
+		if (elfchk_debug&DEBUG_DWARF_CONTENTS) {
+			fprintf(stderr," r%d", tmp );
+                }
 
 		/* Operand 2 - ULEB128 register */
 		tmp = decode_uleb128(ptr,&numused);
                 ptr += numused;
                 used += numused;
+		if (elfchk_debug&DEBUG_DWARF_CONTENTS) {
+			fprintf(stderr," r%d\n", tmp );
+                }
 		break;
 
 /* 	case DW_CFA_expression: */
@@ -390,21 +464,36 @@ dumpbytes(ptr,8);
 
 }
 
-int check_FDE(unsigned char *ptr, int *error)
+int check_FDE(CIEFrame *cie_list, unsigned char *ptr,  int *error)
 {
-        int	used=0;
+        int	numused,used=0;
         FDEFrameHeader	fdeimage;
+		CIEFrame *cie = 0;
+        unsigned char *offsetptr;
         unsigned char *endptr;
+		unsigned long length;
+        void *fdeimage_ptr;
 
-/*
-fprintf(stderr,"FDE found at %x\n", ptr );
-dumpbytes(ptr,length);
-*/
+        fdeimage_ptr = &fdeimage;
 
-        fdeimage.length = *(int *)ptr;
-        ptr += 4;
-        used += 4;
+        memset(fdeimage_ptr,0,sizeof(FDEFrameHeader));
+
+	    length = *(unsigned int *)ptr;
+		ptr += 4;
+		used +=4;
+
+		if (length == 0xffffffff) {
+		   length = *(unsigned long long *)ptr;
+		   ptr += 8;
+		   used += 8;
+		}
+
+		fdeimage.length = length;
         endptr = ptr + fdeimage.length;
+/*
+fprintf(stderr,"FDE found at %p\n", ptr );
+dumpbytes(ptr,fdeimage.length);
+*/
 
         if (fdeimage.length==0)
         {
@@ -412,24 +501,52 @@ dumpbytes(ptr,length);
                 return 0;
         }
 
-        fdeimage.CIE_pointer = *(int *)ptr;
+        fdeimage.CIE_pointer = *(unsigned int *)ptr;
+		offsetptr = ptr;
         ptr += 4;
         used += 4;
+       
+        for (cie=cie_list; cie !=0; cie=cie->next) {
+		  if ((offsetptr - fdeimage.CIE_pointer) == cie->cie_start_addr) 
+		     break;
+		}
 
-        fdeimage.initial_location = *(long **)ptr;
+        if (cie) {
+		   int numused;
+
+		   if (cie->fde_encoding) {
+		    /* encoded value may be relative, depending on encoding bits 0x70 */
+		    fdeimage.initial_location = read_FDE_encoded(ptr, cie->fde_encoding, &numused);
+			ptr += numused;
+			used += numused;
+		   } else {
+             fdeimage.initial_location = *(long **)ptr; 
+             ptr += sizeof(long);
+             used += sizeof(long);
+		   }
+		} else {
+            fprintf(stderr,"FDE contains invalid CIE pointer %lx\n", 
+			                (unsigned long)fdeimage.CIE_pointer);
+		}
+        
+        /* address_range is always specified as an absolute value */
+        fdeimage.address_range = (long)(*ptr);
         ptr += sizeof(long);
         used += sizeof(long);
-        
-        fdeimage.address_range = *(long *)ptr;
-        ptr += sizeof(long);
-        used += sizeof(long);
-        
-/*         fdeimage.address_range = decode_sleb128(ptr,&numused); */
-/*         ptr += numused; */
-/*         used += 4; */
+
+        if (cie && (cie->augmentation[0] == 'z')) {
+		   fdeimage.augmentation_len = decode_uleb128(ptr,&numused);
+		   ptr+= numused;
+		   used += 4;
+		}
+		if (fdeimage.augmentation_len) {
+		   fdeimage.augmentation_data = ptr;
+		   ptr += fdeimage.augmentation_len;
+		}
 
         if(elfchk_debug & DEBUG_DWARF_CONTENTS) {
-                fprintf(stderr,"length: %x\n", fdeimage.length);
+		        int i;
+                fprintf(stderr,"length: %lx\n", fdeimage.length);
                 fprintf(stderr,"CIE_pointer: %x\n", fdeimage.CIE_pointer);
                 fprintf(stderr,"initial_location: %p\n", 
                         fdeimage.initial_location);
@@ -437,10 +554,16 @@ dumpbytes(ptr,length);
                         fdeimage.address_range,
                         ((char *)fdeimage.initial_location) 
                         + fdeimage.address_range);
-	}
+                fprintf(stderr,"augmentation length: %x\n", fdeimage.augmentation_len);
+                fprintf(stderr,"augmentation data: ");
+				for (i=0; i < fdeimage.augmentation_len; i++) {
+				   fprintf (stderr, " %02x", fdeimage.augmentation_data[i]);
+				}
+                fprintf(stderr,"\n");
+        }
 
         while(ptr<endptr)
-                ptr += check_CFI(ptr, error);
+                ptr += check_CFI(ptr, error, cie->fde_encoding);
 
         return fdeimage.length + 4; /* length+sizeof(length) */
 }
@@ -449,9 +572,9 @@ int check_CFInformation(unsigned char *ptr, int *error)
 {
         int     used, numused = 0;
         CIEFrameImage *frameimg;
-
         *error = 0;
         frameimg = (CIEFrameImage *)ptr;
+		CIEFrame *cie_list = 0;
 
         while (frameimg->length!=0)
         {
@@ -465,7 +588,12 @@ int check_CFInformation(unsigned char *ptr, int *error)
                                 fprintf(stderr, "-----------------------\n");
                                 fprintf(stderr, "CIE record\n");
                         }
-                        used = check_CIE(ptr, error);
+                        used = check_CIE(&cie_list, ptr, error);
+						if (used == -1)  /* alloc failure */
+						{
+						  return -1;
+						}
+						   
                 }
                 else
                 {
@@ -474,7 +602,11 @@ int check_CFInformation(unsigned char *ptr, int *error)
                                 fprintf(stderr, "-----------------------\n");
                                 fprintf(stderr, "FDE record\n");
                         }
-                        used = check_FDE(ptr, error);
+                        used = check_FDE(cie_list, ptr, error); 
+						if (used == -1)  
+						{
+						  return -1;
+						}
                 }
                 numused+=used;
                 ptr += used;
@@ -483,88 +615,157 @@ int check_CFInformation(unsigned char *ptr, int *error)
         return numused;
 }
 
-int check_CIE(unsigned char *ptr, int *error)
+int check_CIE(CIEFrame **cie_list, unsigned char *ptr, int *error)
 {
-        CIEFrameImage *frameimg;
-        CIEFrame frame;
+        /* CIEFrameImage *frameimg; */
+        CIEFrame *frame;
         unsigned char *start = ptr;
+        unsigned long length;
         unsigned char *cfi_start;
         int numused;
 
-        frameimg = (CIEFrameImage *)ptr;
-        frame.length = frameimg->length;
-        frame.cie = frameimg->cie;
-        frame.version = frameimg->version;
-        frame.augmentation = frameimg->augmentation;
+        if ((frame=(CIEFrame *)calloc(1, sizeof(CIEFrame))) == NULL) {
+		   fprintf (stderr, "Unable to alloc CIE frame memory for %08lx\n", (unsigned long)ptr);
+		   return -1; 
+	}
+	frame->next = *cie_list;
+	*cie_list = frame;
+	frame->cie_start_addr = ptr;
+	/* Frameimg will not hold good if first 4 bytes are 0xffffffff */
+	/*
+	frameimg = (CIEFrameImage *)ptr;
+	frame->length = frameimg->length;
+	frame->cie = frameimg->cie;
+	frame->version = frameimg->version;
+	frame->augmentation = frameimg->augmentation;
+	*/
+	length = *(unsigned int *)ptr;
+	ptr += 4;
+
+	if (length == 0xffffffff) {
+		length = *(unsigned long long *)ptr;
+		ptr += 8;
+	}
+		
+	frame->length = length;
+
+	frame->cie = *(unsigned int *)ptr;
+	ptr += 4;
+
+	frame->version = *(unsigned int *)ptr;
+	ptr++;
+
+        frame->augmentation = (char *)ptr;
 
         if (elfchk_debug & DEBUG_DWARF_CONTENTS) {
-                fprintf(stderr,"length: %x\n", frame.length);
-                fprintf(stderr,"cie: %x\n", frame.cie);
-                fprintf(stderr,"ver: %x\n", frame.version);
-                fprintf(stderr,"aug: %s\n", frame.augmentation);
+                fprintf(stderr,"CIE address: %08lx \n ",
+			(unsigned long)(frame->cie_start_addr));
+                fprintf(stderr,"length: %lx\n", frame->length);
+                fprintf(stderr,"cie: %x\n", frame->cie);
+                fprintf(stderr,"ver: %x\n", frame->version);
+                fprintf(stderr,"aug: %s\n", frame->augmentation);
 	}
 
-        ptr = (unsigned char *)(frameimg->augmentation)
-                + strlen(frameimg->augmentation) + 1;
+        ptr = (unsigned char *)(frame->augmentation)
+                + strlen(frame->augmentation) + 1;
 
-        if (strcmp(frame.augmentation, "eh")==0)
-        {
-                ptr += sizeof(long);
-        }
-
-        frame.code_alignment_factor = decode_uleb128(ptr, &numused);
+        frame->code_alignment_factor = decode_uleb128(ptr, &numused);
         ptr += numused;
 
-        frame.data_alignment_factor = decode_sleb128(ptr, &numused);
+        frame->data_alignment_factor = decode_sleb128(ptr, &numused);
         ptr += numused;
 
-        frame.return_address_register = *ptr++;
+        frame->return_address_register = *ptr++;
 
         if(elfchk_debug & DEBUG_DWARF_CONTENTS) {
-                fprintf(stderr,"code_af: %d\n", frame.code_alignment_factor);
-                fprintf(stderr,"data_af: %d\n", frame.data_alignment_factor);
+                fprintf(stderr,"code_af: %d\n", frame->code_alignment_factor);
+                fprintf(stderr,"data_af: %d\n", frame->data_alignment_factor);
                 fprintf(stderr,"return_address_register: %x\n",
-                        frame.return_address_register);
+                        frame->return_address_register);
 	}
 
 
         /* Get the info related with 'z' */
-        if (frame.augmentation[0] == 'z')
+        if (frame->augmentation[0] == 'z')
         {
-                frame.augmentation_len = decode_uleb128(ptr,&numused);
+		        int aug_char;
+
+                frame->augmentation_len = decode_uleb128(ptr,&numused);
                 ptr += numused;
                 if (elfchk_debug & DEBUG_DWARF_CONTENTS) {
                         fprintf(stderr,"augmentation_len: %x\n", 
-                                frame.augmentation_len);
+                                frame->augmentation_len);
                 }
-                cfi_start = ptr + frame.augmentation_len;
+                cfi_start = ptr + frame->augmentation_len;
 
-                /* Get the info related with 'P' */
-                frame.encoding = *ptr++;
-                frame.personality_routine = *(unsigned char **)ptr;
-                ptr += sizeof(unsigned char *);
+                /* Skip 'z'; get the info related with 'P', 'L' and/or 'R' */
+				for (aug_char=1; aug_char < strlen(frame->augmentation); aug_char++)
+				{
+
+				   switch (frame->augmentation[aug_char]) {
+				      case 'P':
+                          frame->per_encoding = *ptr++;
+						  /* get the per routine according to the encoding */
+						  if (frame->per_encoding) {
+                             frame->personality_routine = 
+							       read_FDE_encoded(ptr, frame->per_encoding, &numused);
+						     ptr += numused;
+						  } else {
+                             frame->personality_routine = *(unsigned char **)ptr;
+                             ptr += sizeof(unsigned char *);
+						  }
+						  break;
+					  case 'L':
+						  frame->lsda_encoding = *ptr++;
+						  break;
+					  case 'R':
+						  frame->fde_encoding = *ptr++;
+						  break;
+					  default:
+                          fprintf(stderr,"Unexpected augmentation string %c\n", 
+						                 frame->augmentation[aug_char]);
+						  break;
+				   }
+				   
+				}
 
                 if (elfchk_debug & DEBUG_DWARF_CONTENTS) {
-                        fprintf(stderr,"encoding: %x\n", frame.encoding);
+				     if (frame->per_encoding) {
+                        fprintf(stderr,"per encoding: %x\n", frame->per_encoding);
                         fprintf(stderr,"per routine: %p\n", 
-                                frame.personality_routine);
-                }
+                                frame->personality_routine);
+					 }
+				     if (frame->lsda_encoding) {
+                        fprintf(stderr,"lsda encoding: %x\n", frame->lsda_encoding);
+					 }
+				     if (frame->fde_encoding) {
+                        fprintf(stderr,"fde encoding: %x\n", frame->fde_encoding);
+					 }
+                 }
+
                 /* We use the offset from the augmentation
-                   in this case because we don't know how
-                   to handle the info with L in the augmentation */
+                   to ignore any other information in the augmentation */
                 ptr = cfi_start;
         }
+		else if (strlen(frame->augmentation) != 0)
+		{ /* We ignore the remaining CIE and FDE information as we don't
+				  know how it may be structured */
+            fprintf(stderr,"CIE contains unexpected augmentation string %s\n", frame->augmentation);
+			*error = 1;
+			return -1;
+		}
 
 
         /* Get the CFIs */
+/*
         if (elfchk_debug&DEBUG_DWARF_CONTENTS) {
                 fprintf(stderr,"%x bytes of CFI\n", 
                         (int)(ptr - (unsigned char *)frameimg));
 	}
+*/
 
-        while ( (ptr - (unsigned char *)frameimg) < 
-                frameimg->length + sizeof(int) ) {
-                ptr += check_CFI(ptr, error);
+        while ( (ptr - start) < frame->length + sizeof(int) ) {
+                ptr += check_CFI(ptr, error,/* ptr_encoding */ 0);
 	}
 
         return ptr - start;
@@ -575,6 +776,7 @@ read_FDE_encoded(unsigned char *ptr, unsigned char encoding, int *numused)
 {
         void *tmp;
 
+        *numused = 0;
         if (encoding == DW_EH_PE_omit) 
         {
                 *numused = 0;
@@ -582,33 +784,33 @@ read_FDE_encoded(unsigned char *ptr, unsigned char encoding, int *numused)
 	}
 
         switch (encoding & 0x0f) {
-	case DW_EH_PE_uleb128:
+	case DW_EH_PE_uleb128: /* 0x01 */
 		tmp = (void *) decode_uleb128(ptr,numused);
 		return tmp;
-	case DW_EH_PE_udata2:
+	case DW_EH_PE_udata2: /* 0x02 */
 		tmp = (void *)(long)(*(unsigned short *)ptr);
 		*numused = 2;
 		return tmp;
-	case DW_EH_PE_udata4:
+	case DW_EH_PE_udata4: /* 0x03 */
 		tmp = (void *)(long)(*(unsigned int *)ptr);
 		*numused = 4;
 		return tmp;
-	case DW_EH_PE_udata8:
+	case DW_EH_PE_udata8: /* 0x04 */
 		tmp = (void *)(long)(*(unsigned long long *)ptr);
 		*numused = 8;
 		return tmp;
-	case DW_EH_PE_sleb128:
+	case DW_EH_PE_sleb128: /* 0x09 */
 		tmp = (void *)decode_sleb128(ptr,numused);
 		return tmp;
-	case DW_EH_PE_sdata2:
+	case DW_EH_PE_sdata2: /* 0x0A */
 		tmp = (void *)(long)(*(short *)ptr);
 		*numused = 2;
 		return tmp;
-	case DW_EH_PE_sdata4:
+	case DW_EH_PE_sdata4: /* 0x0B */
 		tmp = (void *)(long)(*(int *)ptr);
 		*numused = 4;
 		return tmp;
-	case DW_EH_PE_sdata8:
+	case DW_EH_PE_sdata8: /* 0x0C */
 		tmp = (void *)(long)(*(long long *)ptr);
 		*numused = 8;
 		return tmp;
